<!-- CSS for the report -->
<link rel="stylesheet" href="https://unpkg.com/purecss@1.0.0/build/pure-min.css" integrity="sha384-nn4HPE8lTHyVtfCBi5yW9d20FjT8BJwUXyWZT9InLYax14RDjBj46LmSztkmNP9w" crossorigin="anonymous">
<link rel="stylesheet" href="https://unpkg.com/purecss@1.0.0/build/grids-responsive-min.css">
<style type="text/css">

/* Add a gutter to Pure's Columns */
.pure-g > div {
  -webkit-box-sizing: border-box;
  -moz-box-sizing: border-box;
  box-sizing: border-box;
}

.pure-g > div {
  padding: 0 0.5em;
}

/*!
Pure-button class copied from pure-min.css,
in case reports are displayed offline.

Pure v1.0.0
Copyright 2013 Yahoo!
Licensed under the BSD License.
https://github.com/yahoo/pure/blob/master/LICENSE.md
*/

.pure-button {
    /* Structure */
    display: inline-block;
    zoom: 1;
    line-height: normal;
    white-space: nowrap;
    vertical-align: middle;
    text-align: center;
    cursor: pointer;
    -webkit-user-drag: none;
    -webkit-user-select: none;
       -moz-user-select: none;
        -ms-user-select: none;
            user-select: none;
    box-sizing: border-box;
}

.pure-button {
    /* Button styling */
    font-family: inherit;
    font-size: 100%;
    padding: 0.5em 1em;
    color: #444; /* rgba not supported (IE 8) */
    color: rgba(0, 0, 0, 0.80); /* rgba supported */
    border: 1px solid #999;  /*IE 6/7/8*/
    border: none rgba(0, 0, 0, 0);  /*IE9 + everything else*/
    background-color: #E6E6E6;
    text-decoration: none;
    border-radius: 2px;
}

/*!
End pure-button class definition from pure-min.css
*/

div.nilearn_report {
    padding-top: 0px;
    border: solid rgb(150, 150, 150);
    border-width: 1pt;
    border-radius: 6pt;
    overflow: hidden;  /* Needed to avoid scrolling in jupyter :( */
    color: black;
    background-color: white;
}

.terminal div.nilearn_report {
    max-width: min(50ex, 100vw); /* Needed in vscode */
}

/* Isolate us a bit from the styles specified by other stylesheets */

div.nilearn_report h1 {
    text-align: left;
    margin-left: 0pt;
    margin-right: 0pt;
    padding-top: 5pt;
    font-size: 1.8em;
    padding-left: 5pt;
}

div.nilearn_report h1:first-child {
    margin-top: 0pt;
    background-color: #f5f5f5;
    border-radius: 6pt 6pt 0pt 0pt;
    padding-left: 5pt;
}

div.nilearn_report summary::-webkit-details-marker {
    margin-right: 2px;
}

div.nilearn_report summary:focus {
    outline-style: none;
}

div.nilearn_report table {
    max-width: 100%;
    border-collapse: collapse;
    margin:50px auto;
    float: right;
}

div.nilearn_report thead th {
  text-align: center;
}

div.nilearn_report tbody tr:nth-child(even) {
   background-color: #ddd;
}

div.nilearn_report details {
    overflow-x: visible;
}

div.nilearn_report div.raise {
    z-index: 1;

}

/* Tooltip container */
div.nilearn_report .withtooltip {
  position: relative;
  border-bottom: 1px dotted black; /* If you want dots under the hoverable text */
}

/* Tooltip text */
div.nilearn_report .withtooltip .tooltiptext {
  visibility: hidden;
  max-width: 35em;
  width: 90%;
  background-color: rgba(24,24,24,0.9);
  color: #fff;
  text-align: left;
  font-size: 12pt;
  font-weight: 200;
  border-radius: 6px;
  padding: 5px;css=resource_path.joinpath('css'),

  margin-left: 3pt;

  /* Position the tooltip text - see examples below! */
  position: absolute;
  z-index: 2;

  width: 99%;
  left: 2pt;
  top: 133%;
}

/* Show the tooltip text when you mouse over the tooltip container */
div.nilearn_report .withtooltip:hover .tooltiptext {
  visibility: visible;
}

div.nilearn_report .image {
  position: relative;
  height: auto;
  width: 100%;
}

div.nilearn_report .image .overlay {
  position: absolute;
  top: 0;
  left: 0;
  opacity: 0;
  transition: .5s ease;
  height: auto;
  width: 100%;
  -webkit-transition: .5s ease;
}

div.nilearn_report .image:hover .overlay {
  opacity: 1;
}

.elem-warn {
    color: #FF0000;
    font-weight: bold;
}

</style>
<div class="nilearn_report">
  <h1 class="withtooltip">
    {{title}}
    <span class="tooltiptext">{{docstring}}</span>
  </h1>
  <div class="pure-g">
  <div class="pure-u-1 pure-u-md-2-3">
      <div class="image">
      <img class="pure-img" width="100%" src="data:image/svg+xml;base64,{{content}}" alt="image"/>
      {{if overlay}}
      <div class="overlay">
        <img class="pure-img" width="100%" src="data:image/svg+xml;base64,{{overlay}}" alt="overlay"/>
      </div>
      {{endif}}
      </div>
  </div>
  <div class="pure-u-1 pure-u-md-1-3 raise">
    {{if warning_message}}
      <p class="elem-warn">{{warning_message}}</p>
    {{endif}}
    {{if description}}
      <p class="elem-desc">{{description}}</p>
    {{endif}}
    <p></p>
    <details>
      <summary class="pure-button">Parameters</summary>
      <table class="pure-table">
        <thead>
          <tr>
            <th>Parameter</th>
            <th>Value</th>
          </tr>
        </thead>

        <tbody>
          {{py: params = parameters.items()}}
          {{for p, v in params}}
            <tr>
              <td data-column="Parameter">{{p}}</td>
              <td data-column="Value">{{v}}</td>
            </tr>
          {{endfor}}
        </tbody>
      </table>

    </details>
    </div>
  </div>
</div>
